import {
    CustomViewInterface,
    FeatureBlockInterface,
    FeatureViewInterface
} from "../RcsbFvSequence/SequenceViews/CustomView/CustomView";
import {
    RcsbFv,
    RcsbFvDisplayTypes,
    RcsbFvRowConfigInterface,
    RcsbFvTrackDataElementInterface
} from "@rcsb/rcsb-saguaro";
import { RegionSelectionInterface } from "../RcsbFvState/RcsbFvSelectorManager";
import {
    StructureViewerPublicInterface, SaguaroRegionList
} from "../RcsbFvStructure/StructureViewerInterface";
import {
    LoadMethod,
    LoadMolstarInterface,
    LoadMolstarReturnType
} from "../RcsbFvStructure/StructureViewers/MolstarViewer/MolstarActionManager";
import {RcsbFvStateManager} from "../RcsbFvState/RcsbFvStateManager";

import {RcsbFvStructureConfigInterface} from "../RcsbFvStructure/RcsbFvStructure";
import {ViewerProps} from "@rcsb/rcsb-molstar/build/src/viewer";
import { updateSiteColors } from "./UniTmpColor";
import { fetchHtpDescriptor, fetchPdbtmJsvLibDescriptor, htpDescriptorToTrackData, jsvLibDescriptorToTrackData } from "./UniTmpHelper";
import { TmDetRcsbPreset } from "./TmTrajectoryHierarchyPreset";
import { TmDetDescriptorCache } from "./tmdet-extension/prop";

export type TmFvConfigParams = {
    pdbEntry: string,
    pdbtmEntry: string,
    htpEntry: string,
    labelAsymId: string,
    side1: string,
    title: string,
    subtitle?: string,
    fvConfigData?: any
};
export async function createFeatureViewerConfing(params: TmFvConfigParams) {

    const modelId = `${params.pdbEntry}_model`;

    const rowConfig: Array<RcsbFvRowConfigInterface> = [
        {
            trackId: "sequenceTrack",
            trackHeight: 20,
            trackColor: "#F9F9F9",
            displayType: RcsbFvDisplayTypes.SEQUENCE,
            nonEmptyDisplay: true,
            rowTitle: "N/A",
            trackData: [
                {
                    begin: 1,
                    value: "N/A"
                }
            ]
        }
    ];

    const fvConfig: FeatureViewInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType> =
        params.fvConfigData
        ?? {
            boardId: `${params.pdbEntry}_board`,
            boardConfig: {
                rowTitleWidth: 190,
                includeAxis: true
            },
            rowConfig: rowConfig,
            sequenceSelectionChangeCallback: () => {},
            sequenceElementClickCallback: () => {},
            sequenceHoverCallback: () => {},
            structureSelectionCallback: () => {},
            structureHoverCallback: () => {}
        };

    fvConfig.sequenceSelectionChangeCallback = (plugin: StructureViewerPublicInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, stateManager: RcsbFvStateManager, sequenceRegion: Array<RcsbFvTrackDataElementInterface>) => {
        stateManager.selectionState.clearSelection("select", {modelId: modelId, labelAsymId: params.labelAsymId});
        if(sequenceRegion.length > 0) {
            const regions = sequenceRegion.map(r => ({
                modelId: modelId,
                labelAsymId: params.labelAsymId,
                region: {begin: r.begin, end: r.end ?? r.begin, source: "sequence"} as RegionSelectionInterface
            }));
            stateManager.selectionState.addSelectionFromMultipleRegions(regions, "select");
            plugin.select(regions.map(r => ({
                ...r,
                begin: r.region.begin,
                end: r.region.end
            })), "select", "set");
        }else{
            plugin.clearSelection("select", {modelId: modelId, labelAsymId: params.labelAsymId})
            plugin.resetCamera();
        }
    };
    fvConfig.sequenceElementClickCallback = (plugin: StructureViewerPublicInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, stateManager: RcsbFvStateManager, d: RcsbFvTrackDataElementInterface) => {
        if(d!=null)
            plugin.cameraFocus(modelId, params.labelAsymId, d.begin, d.end ?? d.begin);
    };
    fvConfig.sequenceHoverCallback = (plugin: StructureViewerPublicInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, stateManager: RcsbFvStateManager, elements: Array<RcsbFvTrackDataElementInterface>) => {
        if(elements == null || elements.length == 0)
            plugin.clearSelection("hover");
        else {
            const e = elements[0];
            plugin.select(elements.map(e=>({
                modelId: modelId,
                labelAsymId: params.labelAsymId,
                begin: e.begin,
                end: e.end ?? e.begin
            })), "hover", "set");
        }
    };
    fvConfig.structureSelectionCallback = (plugin: StructureViewerPublicInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, pfv: RcsbFv, stateManager: RcsbFvStateManager) => {
        const sel: SaguaroRegionList | undefined = stateManager.selectionState.getSelectionWithCondition(modelId, params.labelAsymId, "select");
        if(sel == null) {
            pfv.clearSelection("select");
            plugin.resetCamera();
        }else {
            pfv.setSelection({elements: sel.regions, mode: "select"});
        }
    };
    fvConfig.structureHoverCallback = (plugin: StructureViewerPublicInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType>, pfv: RcsbFv, stateManager: RcsbFvStateManager) => {
        const sel: SaguaroRegionList | undefined = stateManager.selectionState.getSelectionWithCondition(modelId, params.labelAsymId, "hover");
        if(sel == null)
            pfv.clearSelection("hover");
        else
            pfv.setSelection({elements:sel.regions, mode:"hover"});
    };

    const block: FeatureBlockInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType> = {
        blockId: `${params.pdbEntry}_${params.labelAsymId}_block_id`,
        featureViewConfig: [fvConfig]
    };

    const customConfig: CustomViewInterface<LoadMolstarInterface<unknown,unknown>,LoadMolstarReturnType> = {
        blockConfig:[block]
    }

    const sequenceConfig = {
        title: params.title,
        subtitle: params.subtitle ?? '',
        config: customConfig
    };

    const molstarConfig: RcsbFvStructureConfigInterface<LoadMolstarInterface<unknown,unknown>,{viewerProps:Partial<ViewerProps>}> = {
        loadConfig: {
            loadMethod: LoadMethod.loadStructureFromUrl,
            loadParams: {
                url: `https://www.ebi.ac.uk/pdbe/entry-files/download/${params.pdbEntry}_updated.cif`,
                format: 'mmcif',
                isBinary: false,
                id: `${params.pdbEntry}_model`,
                reprProvider: TmDetRcsbPreset
            }
        },
        structureViewerConfig: {
            viewerProps:{
                showImportControls: true,
                showSessionControls: false
            }
        }
    };

    updateSiteColors(params.side1 as any);
    if (!params.fvConfigData) {
        const htpDescriptor = await fetchHtpDescriptor(params.htpEntry);
        const jsvlibDescriptor = await fetchPdbtmJsvLibDescriptor(params.pdbtmEntry);

        let sequence = htpDescriptor.Sequence.Seq;
        sequence = sequence.replace(/\s+/g, '');
    
        fvConfig.boardConfig.length = parseInt(htpDescriptor.Sequence['@attributes'].Length);
        fvConfig.rowConfig[0].rowTitle = `${htpDescriptor['@attributes'].id} sequence`;
        rowConfig[0].trackData![0].value = sequence;
        rowConfig.push(...htpDescriptorToTrackData(htpDescriptor));
    
        const descriptor = TmDetDescriptorCache.get(params.pdbEntry);
        const chain = descriptor?.chains.filter(ch => ch.chain_label == params.labelAsymId)[0];
        const offset = parseInt(chain?.residues[0].pdb_res_label!) - 1;
        jsvlibDescriptor.regions.forEach((region: any) => {
            region.start += offset;
            region.end += offset;
        });
        console.log('JSVLIB:', jsvlibDescriptor);
        rowConfig.push(...jsvLibDescriptorToTrackData(jsvlibDescriptor));
    }

    return { sequenceConfig: sequenceConfig, molstarConfig: molstarConfig };
}
